# Copyright contributors to the Terratorch project
import gc
import importlib

import pytest
import torch

from terratorch.models import EncoderDecoderFactory
from terratorch.models.backbones.prithvi_vit import PRETRAINED_BANDS
from terratorch.models.model import AuxiliaryHead

NUM_CHANNELS = 6
NUM_CLASSES = 10
EXPECTED_SEGMENTATION_OUTPUT_SHAPE = (1, NUM_CLASSES, 224, 224)
EXPECTED_REGRESSION_OUTPUT_SHAPE = (1, 224, 224)
EXPECTED_CLASSIFICATION_OUTPUT_SHAPE = (1, NUM_CLASSES)

PIXELWISE_TASK_EXPECTED_OUTPUT = [
    ("regression", EXPECTED_REGRESSION_OUTPUT_SHAPE),
    ("segmentation", EXPECTED_SEGMENTATION_OUTPUT_SHAPE),
]

VIT_UPERNET_NECK = [
    {"name": "SelectIndices", "indices": [0, 1, 2, 3]},
    {"name": "ReshapeTokensToImage"},
    {"name": "LearnedInterpolateToPyramidal"},
]

PRETRAINED_BANDS = ["RED", "GREEN", "BLUE", "NIR_NARROW", "SWIR_1", "SWIR_2"]

@pytest.fixture(scope="session")
def model_factory() -> EncoderDecoderFactory:
    return EncoderDecoderFactory()


@pytest.fixture(scope="session")
def model_input() -> torch.Tensor:
    return torch.ones((1, NUM_CHANNELS, 224, 224))


backbones = ["convnextv2_base"]
pretrained = [False, True]
@pytest.mark.parametrize("backbone", backbones)
@pytest.mark.parametrize("backbone_pretrained", pretrained)
def test_create_classification_model_resnet(backbone, model_factory: EncoderDecoderFactory, model_input, backbone_pretrained):
    model = model_factory.build_model(
        "classification",
        backbone=backbone,
        decoder="IdentityDecoder",
        backbone_in_chans=len(PRETRAINED_BANDS),
        backbone_pretrained=backbone_pretrained,
        num_classes=NUM_CLASSES,
    )
    model.eval()

    with torch.no_grad():
        assert model(model_input).output.shape == EXPECTED_CLASSIFICATION_OUTPUT_SHAPE

    gc.collect()



# reference torchgeo https://torchgeo.readthedocs.io/en/stable/_modules/torchgeo/models/vit.html

from torchgeo.models.croma import CROMA
import logging
from collections.abc import Callable
from functools import partial
import huggingface_hub
import torch.nn as nn
from typing import List
import huggingface_hub
from torchvision.models._api import Weights, WeightsEnum
import torch
from terratorch.models.backbones.select_patch_embed_weights import select_patch_embed_weights

from terratorch.registry import TERRATORCH_BACKBONE_REGISTRY

class CROMAWrapper(nn.Module):

    def __init__(self, croma_model, croma_meta, weights=None, out_indices=None) -> None:
        """
        Args:
            dofa_model (DOFA): The decoder module to be wrapped.
            weights ()
        """
        super().__init__()
        self.croma_model = croma_model
        self.weights = weights
        self.out_channels = [x['num_chs'] for x in self.croma_model.feature_info]
        self.croma_meta = croma_meta

    def forward(self, x: List[torch.Tensor]) -> torch.Tensor:
        return self.croma_model.forward_intermediates(x, intermediates_only=True)

croma_base_config = {
                      "modalities": ['sar', 'optical'],
                      "encoder_dim": 768,
                      "encoder_depth": 12,
                      "num_heads": 16,
                      "patch_size": 8,
                      "image_size": 120,
                    }

croma_large_config = {
                      "modalities": ['sar', 'optical'],
                      "encoder_dim": 768,
                      "encoder_depth": 12,
                      "num_heads": 16,
                      "patch_size": 8,
                      "image_size": 120,
                    }

@TERRATORCH_BACKBONE_REGISTRY.register
def croma_base(model_bands=None, pretrained: bool = False, ckpt_data: bool = None):
  model = CROMA(**croma_base_config)
  return model 

@TERRATORCH_BACKBONE_REGISTRY.register
def croma_large(model_bands=None, pretrained: bool = False, ckpt_data: str = None):
  model = CROMA(**croma_large_config)
  return model 




import logging

import huggingface_hub
import torch
from torch import nn

from terratorch.registry import TERRATORCH_BACKBONE_REGISTRY

try:
    from galileo.data.earthengine.s1 import S1_BANDS
    from galileo.data.earthengine.s2 import ALL_S2_BANDS
    from galileo.galileo import GalileoWrapper
except:
    logging.getLogger("terratorch").info(
        "The package `galileo` is not installed. If you want to use it, install it using"
        "`pip install git+https://github.com/Joao-L-S-Almeida/terratorch-galileo.git`"
    )


def load_weights(model: nn.Module, ckpt_data: dict, **kwargs) -> nn.Module:
    logging.getLogger("terratorch").info("Loading weights")

    if ckpt_data is not None:
        ckpt_data = huggingface_hub.hf_hub_download(**ckpt_data)

        checkpoint_model = torch.load(ckpt_data, map_location="cpu", weights_only=True)
        state_dict = model.model.state_dict()

        for k in ["head.weight", "head.bias"]:
            if k in checkpoint_model and checkpoint_model[k].shape != state_dict[k].shape:
                logging.info(f"Removing key {k} from pretrained checkpoint")
                del checkpoint_model[k]

        # load pre-trained model
        model.model.load_state_dict(checkpoint_model, strict=False)

    return model


@TERRATORCH_BACKBONE_REGISTRY.register
def galileo_tiny_encoder(
    pretrained: bool = None,
    kind: str = "s1",
    ckpt_data: str | None = None,
    bands: list = None,
    model_bands: list = None,
    **kwargs,
):
    remote_checkpoint_path = {"repo_id": "nasaharvest/galileo", "subfolder": "models/tiny", "filename": "encoder.pt"}

    if not ckpt_data:
        ckpt_data = remote_checkpoint_path

    model = Galileo(kind=kind, bands=bands, model_bands=model_bands, **kwargs)

    if pretrained:
        model = load_weights(model, ckpt_data)

    return model


@TERRATORCH_BACKBONE_REGISTRY.register
def galileo_base_encoder(
    pretrained: bool = None,
    kind: str = "s1",
    ckpt_data: str | None = None,
    bands: list = None,
    model_bands: list = None,
    **kwargs,
):
    remote_checkpoint_path = {"repo_id": "nasaharvest/galileo", "subfolder": "models/base", "filename": "encoder.pt"}

    if not ckpt_data:
        ckpt_data = remote_checkpoint_path

    model = Galileo(kind=kind, bands=bands, model_bands=model_bands, **kwargs)

    if pretrained:
        model = load_weights(model, ckpt_data)

    return model


@TERRATORCH_BACKBONE_REGISTRY.register
def galileo_nano_encoder(
    pretrained: bool = None,
    kind: str = "s1",
    ckpt_data: str | None = None,
    bands: list = None,
    model_bands: list = None,
    **kwargs,
):
    remote_checkpoint_path = {"repo_id": "nasaharvest/galileo", "subfolder": "models/nano", "filename": "encoder.pt"}

    if not ckpt_data:
        ckpt_data = remote_checkpoint_path

    model = Galileo(kind=kind, bands=bands, model_bands=model_bands, **kwargs)

    if pretrained:
        model = load_weights(model, ckpt_data)

    return model


# Built-in wrapper for Galileo FM
class Galileo(nn.Module):
    def __init__(
        self, kind: str = "s1", transpose: bool = False, bands: list = None, model_bands: list = None, **kwargs
    ):
        super().__init__()

        self.kind = kind

        if transpose:
            self.modify_batch = self.modify_batch_
        else:
            self.modify_batch = lambda x: x

        if self.kind == "s1":
            self.model_bands_default = S1_BANDS
        else:
            self.model_bands_default = ALL_S2_BANDS

        self.model_bands = model_bands or self.model_bands_default
        self.bands = bands or self.model_bands_default

        self.model = GalileoWrapper(**kwargs)

        self.out_channels = [self.model.encoder.embedding_size]

    def modify_batch_(self, x):
        dims = x.shape

        if len(dims) == 5:
            return x.permute(0, 2, 3, 4, 1)
        elif len(dims) == 4:
            return x.permute(0, 2, 3, 1)

    def prepare_input_tensor(self, x):
        x = self.modify_batch(x)

        dims = x.shape[:-1] + (len(self.model_bands_default),)
        x_ext = torch.zeros(*dims).to(x.device)

        for j, band in enumerate(self.model_bands):
            i = self.bands.index(band)

            x_ext[..., i] = x[..., j]

        return x_ext

    def forward(self, x):
        x = self.prepare_input_tensor(x)

        output = self.model(**{f"{self.kind}": x})

        return [output]



# Copyright contributors to the Terratorch project

"""Command-line interface to TerraTorch."""

import logging
import sys

from huggingface_hub import hf_hub_download

from terratorch.cli_tools import build_lightning_cli

try:
    from benchmark.main import main as iterate_main

    TERRATORCH_ITERATE_INSTALLED = True

except ImportError:
    TERRATORCH_ITERATE_INSTALLED = False


def main():
    if len(sys.argv) == 1:
        print(
            "usage: terratorch [-h] [-c CONFIG] [--print_config[=flags]] {fit,validate,test,predict,compute_statistics,init,iterate} ..."
        )
        exit(0)
    if len(sys.argv) >= 2 and sys.argv[1] == "iterate":
        # if user runs "terratorch iterate" and terratorch-iterate has not been installed
        if not TERRATORCH_ITERATE_INSTALLED:
            print(
                "Error! terratorch-iterate has not been installed. If you want to install it,"
                "run 'pip install terratorch-iterate'"
            )
        # if user runs "terratorch iterate" and terratorch-iterate has been installed
        else:
            # delete iterate argument
            del sys.argv[1]
            iterate_main()
    elif sys.argv[1] == "init":
        logger = logging.getLogger("terratorch-init")
        logger.info("Initializing TerraTorch...")
        logger.info("Downloading model congfigs...")
        files_to_download = [
            {
                "repo_id": "ibm-granite/granite-geospatial-biomass",
            },
            {
                "repo_id": "ibm-granite/granite-geospatial-wxc-downscaling",
            },
            {
                "repo_id": "ibm-granite/granite-geospatial-canopyheight",
            },
            {
                "repo_id": "ibm-granite/granite-geospatial-land-surface-temperature",
            },
            {
                "repo_id": "ibm-granite/granite-geospatial-uki",
            },
            {
                "repo_id": "ibm-granite/granite-geospatial-uki-flooddetection",
            },
            {
                "repo_id": "ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
            },
            {
                "repo_id": "ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M-rollout",
            },
            {
                "repo_id": "ibm-nasa-geospatial/Prithvi-WxC-1.0-2300m-gravity-wave-parameterization",
            },
            {
                "repo_id": "ibm-nasa-geospatial/Prithvi-EO-2.0-300M-TL",
                "filename": "config.json",
            },
            {
                "repo_id": "ibm-nasa-geospatial/Prithvi-EO-2.0-600M-TL",
                "filename": "config.json",
            },
            {
                "repo_id": "ibm-nasa-geospatial/Prithvi-EO-2.0-300M-BurnScars",
                "filename": "burn_scars_config.yaml",
            },
            {
                "repo_id": "ibm-nasa-geospatial/Prithvi-EO-1.0-100M-burn-scar",
            },
            {
                "repo_id": "ibm-nasa-geospatial/Prithvi-EO-1.0-100M-multi-temporal-crop-classification",
            },
            {
                "repo_id": "ibm-nasa-geospatial/Prithvi-EO-2.0-300M",
                "filename": "config.json",
            },
            {
                "repo_id": "ibm-nasa-geospatial/Prithvi-EO-1.0-100M",
            },
            {
                "repo_id": "ibm-nasa-geospatial/Prithvi-EO-2.0-300M-TL-Sen1Floods11",
            },
            {
                "repo_id": "ibm-nasa-geospatial/Prithvi-EO-1.0-100M-sen1floods11",
            },
        ]

        for config in files_to_download:
            file_path = hf_hub_download(
                repo_id=config["repo_id"],
                filename=config.get("filename", "") or "config.yaml",
            )
            logger.info(f"File downloaded to: {file_path}")
    else:
        _ = build_lightning_cli()


if __name__ == "__main__":
    main()



import gc
import pdb

import pytest
import torch
from torch import nn

from terratorch.models.utils import TemporalWrapper
from terratorch.registry import BACKBONE_REGISTRY


# Define a dummy encoder for testing
class DummyEncoder(nn.Module):
    def __init__(self, out_channels=64, in_channels=3):
        super().__init__()
        self.out_channels = out_channels
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=3, padding=1)

    def forward(self, x):
        return self.conv(x)


@pytest.fixture
def dummy_encoder():
    return DummyEncoder()


def test_temporal_wrapper_swin_forward_shapes(dummy_encoder):
    # Built-in Swin
    NUM_CHANNELS = 6
    encoder = BACKBONE_REGISTRY.build("prithvi_swin_B", out_indices=[0, 1, 2, 3])

    wrapper = TemporalWrapper(encoder)
    batch_size = 2

    # Test case 2: Valid input shape
    x = torch.randn(batch_size, NUM_CHANNELS, 4, 224, 224)  # [B, C, T, H, W]
    output = wrapper(x)
    assert [o.shape for o in output] == [
        torch.Size([2, 56, 56, 128]),
        torch.Size([2, 28, 28, 256]),
        torch.Size([2, 14, 14, 512]),
        torch.Size([2, 7, 7, 1024]),
    ]

    # Satlas Swin
    NUM_CHANNELS = 6
    encoder = BACKBONE_REGISTRY.build(
        "satlas_swin_b_sentinel2_si_ms", model_bands=[0, 1, 2, 3, 4, 5], out_indices=[1, 3, 5, 7]
    )

    wrapper = TemporalWrapper(encoder)
    batch_size = 2

    # Test case 2: Valid input shape
    x = torch.randn(batch_size, NUM_CHANNELS, 4, 224, 224)  # [B, C, T, H, W]
    output = wrapper(x)
    assert [o.shape for o in output] == [
        torch.Size([2, 56, 56, 128]),
        torch.Size([2, 28, 28, 256]),
        torch.Size([2, 14, 14, 512]),
        torch.Size([2, 7, 7, 1024]),
    ]

    gc.collect()


def test_temporal_wrapper_initialization(dummy_encoder):
    # Test valid initialization with default parameters
    wrapper = TemporalWrapper(dummy_encoder)
    assert wrapper.pooling_type == "mean"
    assert not wrapper.concat
    assert wrapper.out_channels == dummy_encoder.out_channels

    # Test valid initialization with custom parameters
    wrapper = TemporalWrapper(dummy_encoder, pooling="max", concat=True, n_timestamps=4)
    assert wrapper.pooling_type == "max"
    assert wrapper.concat
    assert wrapper.out_channels == dummy_encoder.out_channels * 4

    # Test invalid pooling type
    with pytest.raises(ValueError, match="Pooling must be 'mean', 'max' or 'diff'"):
        TemporalWrapper(dummy_encoder, pooling="invalid")

    # Test encoder without out_channels attribute
    class InvalidEncoder(nn.Module):
        def forward(self, x):
            return x

    with pytest.raises(AttributeError, match="Encoder must have an `out_channels` attribute"):
        TemporalWrapper(InvalidEncoder())

    gc.collect()


def test_temporal_wrapper_forward_shapes(dummy_encoder):
    wrapper = TemporalWrapper(dummy_encoder)
    batch_size = 2

    # Test case 1: Invalid number of dimensions
    x = torch.randn(batch_size, 3, 32, 32)  # 4D tensor
    with pytest.raises(ValueError, match=r"Expected input shape \[B, C, T, H, W\]"):
        wrapper(x)

    # Test case 2: Valid input shape
    x = torch.randn(batch_size, 3, 4, 32, 32)  # [B, C, T, H, W]
    output = wrapper(x)
    assert isinstance(output, list)
    assert len(output) == 1  # Single feature map
    assert output[0].shape == (batch_size, dummy_encoder.out_channels, 32, 32)

    # Test case 3: Different number of timepoints
    x = torch.randn(batch_size, 3, 6, 32, 32)  # 6 timepoints
    output = wrapper(x)
    assert isinstance(output, list)
    assert len(output) == 1
    assert output[0].shape == (batch_size, dummy_encoder.out_channels, 32, 32)

    # Test case 4: Invalid number of input channels
    x = torch.randn(batch_size, 4, 3, 32, 32)  # 4 channels
    with pytest.raises(RuntimeError):
        wrapper(x)  # Should fail because Conv2d expects 3 channels

    gc.collect()


def test_temporal_wrapper_pooling_modes(dummy_encoder):
    batch_size = 2
    timesteps = 4

    ### conv-like features
    n_channels = 3
    x = torch.randn(batch_size, n_channels, timesteps, 224, 224)
    encoder = BACKBONE_REGISTRY.build("timm_resnet50")

    # Test mean pooling
    wrapper = TemporalWrapper(encoder, pooling="mean")
    output = wrapper(x)
    assert isinstance(output, list)
    assert len(output) == 5
    assert output[0].shape == (batch_size, encoder.out_channels[0], 112, 112)

    gc.collect()
    # Test max pooling
    wrapper = TemporalWrapper(encoder, pooling="max")
    output = wrapper(x)
    assert isinstance(output, list)
    assert len(output) == 5
    assert output[0].shape == (batch_size, encoder.out_channels[0], 112, 112)

    gc.collect()
    # Test concatenation
    wrapper = TemporalWrapper(encoder, concat=True, n_timestamps=timesteps)
    output = wrapper(x)
    assert isinstance(output, list)
    assert len(output) == 5
    # For concatenation, channels should be multiplied by number of timesteps
    assert output[0].shape == (batch_size, encoder.out_channels[0] * timesteps, 112, 112)

    gc.collect()
    # Test diff
    wrapper = TemporalWrapper(encoder, pooling="diff")
    output = wrapper(x[:, :, [0, 1], ...])
    assert isinstance(output, list)
    assert len(output) == 5
    assert output[0].shape == (batch_size, encoder.out_channels[0], 112, 112)

    gc.collect()

    ### transformer-like features
    encoder = BACKBONE_REGISTRY.build("clay_v1_base")
    n_channels = 6
    x = torch.randn(batch_size, n_channels, timesteps, 256, 256)
    n_tokens = 1025
    # Test mean pooling
    wrapper = TemporalWrapper(encoder, pooling="mean")
    output = wrapper(x)
    assert isinstance(output, list)
    assert len(output) == 12
    assert output[0].shape == (batch_size, n_tokens, encoder.out_channels[0])

    gc.collect()
    # Test max pooling
    wrapper = TemporalWrapper(encoder, pooling="max")
    output = wrapper(x)
    assert isinstance(output, list)
    assert len(output) == 12
    assert output[0].shape == (batch_size, n_tokens, encoder.out_channels[0])

    gc.collect()
    # Test concatenation
    wrapper = TemporalWrapper(encoder, concat=True, n_timestamps=timesteps)
    output = wrapper(x)
    assert isinstance(output, list)
    assert len(output) == 12
    # For concatenation, channels should be multiplied by number of timesteps
    assert output[0].shape == (batch_size, n_tokens, encoder.out_channels[0] * timesteps)

    gc.collect()
    # Test diff
    wrapper = TemporalWrapper(encoder, pooling="diff")
    print(x[:, :, [0, 1], ...].shape)
    output = wrapper(x[:, :, [0, 1], ...])
    assert isinstance(output, list)
    assert len(output) == 12
    assert output[0].shape == (batch_size, n_tokens, encoder.out_channels[0])

    gc.collect()

    ### Swin-like features
    encoder = BACKBONE_REGISTRY.build(
        "satlas_swin_b_sentinel2_si_ms", model_bands=[0, 1, 2, 3, 4, 5], out_indices=[1, 3, 5, 7]
    )

    n_channels = 6
    x = torch.randn(batch_size, n_channels, timesteps, 256, 256)
    # pdb.set_trace()
    # Test mean pooling
    wrapper = TemporalWrapper(encoder, pooling="mean", features_permute_op=(0, 3, 1, 2))
    output = wrapper(x)
    assert isinstance(output, list)
    assert len(output) == 4
    assert output[0].shape == (batch_size, 64, 64, encoder.out_channels[0])

    gc.collect()
    # Test max pooling
    wrapper = TemporalWrapper(encoder, pooling="max", features_permute_op=(0, 3, 1, 2))
    output = wrapper(x)
    assert isinstance(output, list)
    assert len(output) == 4
    assert output[0].shape == (batch_size, 64, 64, encoder.out_channels[0])

    gc.collect()
    # Test concatenation
    wrapper = TemporalWrapper(encoder, concat=True, n_timestamps=timesteps, features_permute_op=(0, 3, 1, 2))
    output = wrapper(x)
    assert isinstance(output, list)
    assert len(output) == 4
    # For concatenation, channels should be multiplied by number of timesteps
    assert output[0].shape == (batch_size, 64, 64, encoder.out_channels[0] * timesteps)

    gc.collect()
    # Test diff
    wrapper = TemporalWrapper(encoder, pooling="diff", features_permute_op=(0, 3, 1, 2))
    output = wrapper(x[:, :, [0, 1], ...])
    assert isinstance(output, list)
    assert len(output) == 4
    assert output[0].shape == (batch_size, 64, 64, 128)

    gc.collect()



from torch import nn, Tensor
import torch 
from terratorch.registry import BACKBONE_REGISTRY
import pdb

class DecoderNotFoundError(Exception):
    pass

def extract_prefix_keys(d: dict, prefix: str) -> dict:
    extracted_dict = {}
    remaining_dict = {}
    for k, v in d.items():
        if k.startswith(prefix):
            extracted_dict[k[len(prefix) :]] = v
        else:
            remaining_dict[k] = v

    return extracted_dict, remaining_dict


def pad_images(imgs: Tensor, patch_size: int | list, padding: str) -> Tensor:
    p_t = 1
    if isinstance(patch_size, int):
         p_h = p_w = patch_size
    elif len(patch_size) == 1:
        p_h = p_w = patch_size[0]
    elif len(patch_size) == 2:
        p_h, p_w = patch_size
    elif len(patch_size) == 3:
        p_t, p_h, p_w = patch_size
    else:
        raise ValueError(f'patch size {patch_size} not valid, must be int or list of ints with length 1, 2 or 3.')

    # Double the patch size to ensure the resulting number of patches is divisible by 2 (required for many decoders)
    p_h, p_w = p_h * 2, p_w * 2

    if p_t > 1 and len(imgs.shape) < 5:
        raise ValueError(f"Multi-temporal padding requested (p_t = {p_t}) "
                         f"but no multi-temporal data provided (data shape = {imgs.shape}).")

    h, w = imgs.shape[-2:]
    t = imgs.shape[-3] if len(imgs.shape) > 4 else 1
    t_pad, h_pad, w_pad = (p_t - t % p_t) % p_t, (p_h - h % p_h) % p_h, (p_w - w % p_w) % p_w
    if t_pad > 0:
        # Multi-temporal padding
        imgs = torch.stack([
            nn.functional.pad(img, (0, w_pad, 0, h_pad, 0, t_pad), mode=padding)
            for img in imgs  # Apply per image to avoid NotImplementedError from torch.nn.functional.pad
        ])
    elif h_pad > 0 or w_pad > 0:
        imgs = torch.stack([
            nn.functional.pad(img, (0, w_pad, 0, h_pad), mode=padding)
            for img in imgs  # Apply per image to avoid NotImplementedError from torch.nn.functional.pad
        ])
    return imgs


def _get_backbone(backbone: str | nn.Module, **backbone_kwargs) -> nn.Module:
    use_temporal = backbone_kwargs.pop('use_temporal', None)
    temporal_pooling = backbone_kwargs.pop('temporal_pooling', None)
    concat = backbone_kwargs.pop('temporal_concat', None)
    if isinstance(backbone, nn.Module):
        model = backbone
    else:
        model = BACKBONE_REGISTRY.build(backbone, **backbone_kwargs)

    # Apply TemporalWrapper inside _get_backbone
    if use_temporal:
        model = TemporalWrapper(model, pooling=temporal_pooling, concat=concat)

    return model

def subtract_along_dim2(tensor: torch.Tensor):
    return tensor[:, :, 0, ...] - tensor[:, :, 1, ...]


class TemporalWrapper(nn.Module):
    def __init__(self, encoder: nn.Module, pooling="mean", concat=False, n_timestamps=None, features_permute_op = None):
        """
        Wrapper for applying a temporal encoder across multiple time steps.

        Args:
            encoder (nn.Module): The feature extractor (backbone).
            pooling (str): Type of pooling ('mean', 'max', 'diff') with 'diff' working only with 2 timestamps.
            concat (bool): Whether to concatenate features instead of pooling.
            n_timestamps (int): Number of timestamps. Necessary in case of concat.
            features_permute_op (list): Permutation operation to perform on the features before aggregation. This is in case the features to do not match either 'BCHW' or 'BLC' formats. It is reversed once aggregation has happened.
        """
        super().__init__()
        self.encoder = encoder
        self.concat = concat
        self.pooling_type = pooling
        self.n_timestamps = n_timestamps
        self.features_permute_op = features_permute_op

        if pooling not in ["mean", "max", "diff"]:
            raise ValueError("Pooling must be 'mean', 'max' or 'diff'")

        # Ensure the encoder has an out_channels attribute
        if hasattr(encoder, "out_channels"):
            self.out_channels = encoder.out_channels * (1 if not concat else self.n_timestamps)
        else:
            raise AttributeError("Encoder must have an `out_channels` attribute.")


    def forward(self, x: torch.Tensor) -> list[torch.Tensor]:
        """
        Forward pass for temporal processing.

        Args:
            x (Tensor): Input tensor of shape [B, C, T, H, W].

        Returns:
            List[Tensor]: A list of processed tensors, one per feature map.
        """

        if x.dim() != 5:
            raise ValueError(f"Expected input shape [B, C, T, H, W], but got {x.shape}")

        if (self.pooling_type == 'diff') & (x.shape[2] != 2):
            raise ValueError(f"Expected 2 timestamps for aggregation method 'diff'")

        batch_size, _, timesteps, _, _ = x.shape

        # Initialize lists to store feature maps at each timestamp
        num_feature_maps = None  # Will determine dynamically
        features_per_map = []  # Stores feature maps across timestamps

        for t in range(timesteps):
            feat = self.encoder(x[:, :, t, :, :])  # Extract features at timestamp t
                
            if not isinstance(feat, list):  # If the encoder outputs a single feature map, convert to list
                if isinstance(feat, tuple):
                    feat = list(feat)
                else:
                    feat = [feat]

            if num_feature_maps is None:
                num_feature_maps = len(feat)  # Determine how many feature maps the encoder produces

            for i, feature_map in enumerate(feat):
                if len(features_per_map) <= i:
                    features_per_map.append([])  # Create list for each feature map

                feature_map = feature_map[0] if isinstance(feature_map, tuple) else feature_map
                if self.features_permute_op is not None:
                    if len(self.features_permute_op) != len(feature_map.shape):
                        ValueError(f"Expected features_permute_op to have same number of dimensions of features, but got {len(self.features_permute_op)} and {len(feature_map.shape)}")
                    # print('Old shape:', feature_map.shape)
                    feature_map = torch.permute(feature_map, self.features_permute_op)
                    # print('New shape:', feature_map.shape)
                    
                features_per_map[i].append(feature_map)  # Store feature map at time t
                    
        # Stack features along the temporal dimension
        for i in range(num_feature_maps):
            try:
                features_per_map[i] = torch.stack(features_per_map[i], dim=2)  # Shape: [B, C', T, H', W']
            except RuntimeError as e:
                raise

        # Apply pooling or concatenation
        if self.concat:
            features_per_map_agg = [feat.reshape(batch_size, -1, feat.shape[-2], feat.shape[-1]) if len(feat.shape) == 5 else feat.reshape(batch_size, feat.shape[-3], -1) for feat in features_per_map]
        elif self.pooling_type == "max":
            features_per_map_agg = [torch.max(feat, dim=2)[0] for feat in features_per_map]  # Max pooling across T
        elif self.pooling_type == "diff":
            features_per_map_agg = [feat[:, :, 0, ...] - feat[:, :, 1, ...] for feat in features_per_map]
        else:
            features_per_map_agg = [torch.mean(feat, dim=2) for feat in features_per_map]
        
        if self.features_permute_op is not None:
            # use position in the permutation op as the value of the permuation and the original value of permutation as the index
            reverse_permuation_op = [None] * len(self.features_permute_op)
            for i in range(len(self.features_permute_op)):
                reverse_permuation_op[self.features_permute_op[i]] = i
            features_per_map_agg = [torch.permute(feat, reverse_permuation_op) for feat in features_per_map_agg]

        return features_per_map_agg
